/***************************************************************************
  Arcade machine emulator V0.06 04/17/01

  Z80 engine by Marcel de Kogel
  Emulator by Jim Skilskyj based on the structure
  of the emulators by Brian Peek and Chris Hardy.
  (read the emulators howto to learn more)
  Also some code ideas from Phoenix emulator by Richard Davies

  This emulator is only to guide you on how to use the Z80 emulator 
  and how to write code to emulate an arcade machine.

  Updates:
  v0.01 First release, just happy to get Z80 to work.
  v0.02 Added the background characters (stars etc.) without flicker.
	Cleaned up code.
  v0.03 Now playable. Using Allegro keyboard routines.
  v0.04 Cleaned up ROM load routine
  v0.05 Now decodes graphics in B&W (graphics decode by Peter Dons Tychsen)
  v0.06 Added the color decode routine.

***************************************************************************/

#include <stdio.h>
#include <string.h>
#include <malloc.h>
#include <conio.h>
#include "z80.h"
#include <allegro.h>

//Function Prototypes
int run_machine(void);
int kbd_in(int keycode);
void machine_exit(char *derror);
byte Load_ROMs(char *romfile, word Offset, word Size);
void Video_Decode(void);
void load_color_proms();
void pdt_test_palette();
int Video_Draw(void);
int test_proms(void);
void MakeSoundA(byte V);
void MakeSoundB(byte V);


unsigned char RAM[0x10000];
int DipSwitchSYNC = 0;
FILE *in;
char *derror;

Z80_Regs regs;
int Z80_IRQ=-1;
int Z80_Iperiod = 12000; // for Phoenix

PALETTE palette_332;
RGB black={0,0,0};

unsigned int color_palette[0x100]; 

BITMAP *video;
BITMAP *video_buffer_fg;
BITMAP *video_buffer_bg;
BITMAP *video_buffer_mk;
unsigned int scroll_reg;
int retrace_count;
float retrace_speed;

void main(int argc,char **argv)
{

clrscr();

printf("\nYet Another Phoenix Emulator v0.06 04/17/01\n"
"By Jim Skilskyj\n"
"The keyboard definitions are:\n"
"\n"
" Game keys   3       Insert coin\n"
"             1       Start 1 player game\n"
"             2       Start 2 player game\n"
"             Arrows  Move Around\n"
"             CTL     Fire\n"
"             P       Pause\n"
"             ESC     Exit emulator\n"
"\n"
"Press <ENTER> to continue.\n");

allegro_init();     // allegro init
install_keyboard(); // just init allegro keyboard routines
readkey(); // Pause screen
clrscr();

if (!Load_ROMs("IC45",0x0000,0x0800)) machine_exit(derror = "Missing IC45");
if (!Load_ROMs("IC46",0x0800,0x0800)) machine_exit(derror = "Missing IC46");
if (!Load_ROMs("IC47",0x1000,0x0800)) machine_exit(derror = "Missing IC47");
if (!Load_ROMs("IC48",0x1800,0x0800)) machine_exit(derror = "Missing IC48");
if (!Load_ROMs("IC49",0x2000,0x0800)) machine_exit(derror = "Missing IC49");
if (!Load_ROMs("IC50",0x2800,0x0800)) machine_exit(derror = "Missing IC50");
if (!Load_ROMs("IC51",0x3000,0x0800)) machine_exit(derror = "Missing IC51");
if (!Load_ROMs("IC52",0x3800,0x0800)) machine_exit(derror = "Missing IC52");

//GFX ROMs load
if (!Load_ROMs("IC39",0x8000,0x0800)) machine_exit(derror = "Missing GFX IC39");
if (!Load_ROMs("IC23",0x8800,0x0800)) machine_exit(derror = "Missing GFX IC23");
if (!Load_ROMs("IC40",0x9000,0x0800)) machine_exit(derror = "Missing GFX IC40");
if (!Load_ROMs("IC24",0x9800,0x0800)) machine_exit(derror = "Missing GFX IC24");

// GFX Mode Init
set_color_depth(8);
set_gfx_mode(GFX_MODEX,320,400,320,400);
generate_332_palette(palette_332);
set_palette(palette_332);
set_color(0,&black);
load_color_proms();
Video_Decode();

pdt_test_palette();
readkey();
clear(screen);
test_proms(); 
readkey();

clear(screen);

if (run_machine() != 0)
	{
		printf("Unable to start emulation\n");
		machine_exit(derror = "Z80 Failure");
	}

}

int run_machine(void) // Start the CPU emulation
{
retrace_count=0;
Z80_Reset();
Z80(); 
return (0);
}

int kbd_in(int keycode) // function for allegro keyboard
{
return key[keycode];
}

void machine_exit(char *derror)
{
textmode(C80);
if (derror != "")
{
printf("Emulator Error: %s",derror);
}
remove_keyboard(); 
Z80_Running = 0;
exit(0);
}

byte Load_ROMs(char *romfile, word Offset, word Size)
{
// This loads the ROMs from the file images

in=fopen(romfile, "rb");
if (!in) return(0);
fread(&RAM[Offset], Size, 1 ,in);
fclose (in);
return(1);
}

void Video_Decode(void)
{
// Convert the graphics ROMs to a bitmap format

unsigned int line1,line2,bit1,bit2;
unsigned int i,j,c;
unsigned int lookup=0;

video=create_bitmap_ex(8,0x1000,8);
video_buffer_fg=create_bitmap_ex(8,208,256);
video_buffer_bg=create_bitmap_ex(8,208,256);
video_buffer_mk=create_bitmap_ex(8,208,256);

clear(video);
clear(video_buffer_fg);
clear(video_buffer_bg);
clear(video_buffer_mk);
clear(screen);

for(c=0;c<0x200;c++)     // do a char
	    {
	    for(i=0;i<8;i++) //do a ver. line
	       {
	       line1=RAM[0x8000+(c*8)+7-i];
	       line2=RAM[0x9000+(c*8)+7-i];
		    for(j=0;j<8;j++) //do a hor. line
		       {
		       bit1=((line1&(0x01<<j))>>j);
		       bit2=((line2&(0x01<<j))>>j);
		       lookup=(((c%256)>>5)
			 | (bit1<<3) | (bit2<<4)
			 | (!((c&0xFF00)&&1)<<5));

			  _putpixel(video,((c*8)+i),j,color_palette[lookup]);

		       }
	       }
	    }
}

void load_color_proms()
{
int i=0;
int r=0,g=0,b=0;

unsigned char buffer[0x200];

// The quick and dirty way to load ROMs
in =fopen("IC41_A.BIN", "rb");
if (!in)
{
printf("ROM IC41_A missing!\n");
exit(0);
}
fread(&buffer[0x00], 0x100, 1 ,in);
fclose (in);
i++;

in =fopen("IC40_B.BIN", "rb");
if (!in)
{
printf("ROM IC40_B missing!\n");
exit(0);
}
fread(&buffer[0x100], 0x100, 1 ,in);
fclose (in);

for(i=0;i<128;i++)
	{

	    r=((((buffer[i+0x100]>>0)&0x01))|((((buffer[i]>>0)&0x01))<<1))*85;
	    g=((((buffer[i+0x100]>>2)&0x01))|((((buffer[i]>>2)&0x01))<<1))*85;
	    b=((((buffer[i+0x100]>>1)&0x01))|((((buffer[i]>>1)&0x01))<<1))*85;
	    
	    color_palette[i]=makecol8(r,g,b);

	    if((buffer[i+0x100]==0)&&(buffer[i]==0)) color_palette[i]=0;

	}

	for(i=128;i<256;i++) // The rest are white
	{
	  color_palette[(i)] = makecol8(255,255,255);
	}
}

void pdt_test_palette()
	{
// Used for debugging
    int i=0;
    int j=0;
    int k=0;
     
    for(i=0;i<256;i++)     //do a color
	{
	for (j=0;j<8;j++)  //do a row
		{
	    for(k=0;k<8;k++)  // do a column
			    {
		putpixel(screen,(8*(i%16))+j,(8*(i/16))+k,color_palette[i]);
		putpixel(screen,(8*(i%16))+j,(8*(i/16))+k+160,i);
		}
	    }
	}
    }


int Video_Draw(void)
{
    unsigned int x,y,c,d;
    unsigned int background_y;
    for(y = 0; y < 32; y++)
		{
		int pos1 =  0x4000+32*(26-1)+y;
		int pos2 =  0x4800+32*(26-1)+y;
		for (x = 0; x < 26; x++)
			{
			background_y=((y*8)+(256-scroll_reg))%256;
			c = RAM[pos1];
			d = RAM[pos2];

			// BACKGROUND LAYER WITH WRAP-AROUND
		    blit(video,video_buffer_bg,0x800+(d*8),0,(x*8),
		    background_y,8,8);

	    // FRONT LAYER
	    blit(video,video_buffer_fg,(c*8),0,(x*8),(y*8),8,8);
	    pos1 -= 32; // Foreground
	    pos2 -= 32; // Background
			}
	}

    // UPDATE SCREEN

// foreground only use blit(video_buffer_fg,screen,0,0,0,0,208,256);
 
//  blit(video_buffer_fg,screen,0,0,0,0,208,256);

    blit(video_buffer_bg,video_buffer_mk,0,8,0,8,208,248);
    masked_blit(video_buffer_fg,video_buffer_mk,0,0,0,0,208,256);
    blit(video_buffer_mk,screen,0,0,0,0,208,256);

    //textprintf(screen,font,0,33*8,makecol8(255,255,255),"SCROLL: %d",pdt_scroll_reg);

return(1);
}

int test_proms(void)
{
    int i;
    int j=0;
    int k=0;
    for (i=0;i<0x200;i++)
	{
	k=i/20;
	j=(k*20*8);
	blit(video,screen,(8*i),0,((8*i)-j),(k*8),8,8);
	}
return(1);
}

// z80 emulator functions

unsigned Z80_RDMEM (dword A)
{
// gotoxy(1,34);
// printf("Reading from location %04X\n", A);
 

if(A >= 0x7800 && A <=0x7bff)
	
	{
	if(DipSwitchSYNC == 1)
		{
		Video_Draw();
		DipSwitchSYNC = 0;
		return 0x80;
		}
	return 0x0;
	}
	
else if (A >= 0x7000 && A <= 0x73FF)
	{
	// Reading from buttons port
	// This is where you call the keyboard routine
	
	byte hexdata = 0xff;
	if (kbd_in(1)) machine_exit(derror=""); // Q key = exit
	if (kbd_in(4)) hexdata &= ~(1<<0);  // 3 key = coin 1
	if (kbd_in(2)) hexdata &= ~(1<<1);  // 1 key = 1p start
	if (kbd_in(3)) hexdata &= ~(1<<2);  // 2 key = 2p start
	if (kbd_in(29)) hexdata &= ~(1<<4); // ctl = fire
	if (kbd_in(77)) hexdata &= ~(1<<5); // arrow = right
	if (kbd_in(75)) hexdata &= ~(1<<6); // arrow = left
	if (kbd_in(56)) hexdata &= ~(1<<7); // alt = shield
	return hexdata; 
	}
else return RAM[A];

}

void Z80_WRMEM(dword A,byte V)
{
if (A < 0x4000)
	{
	// Attempting to write to ROM, just return
	return;
	}
	else if(A >= 0x5800 && A <= 0x5bff) scroll_reg = V;
	else if(A > 0x3fff && A < 0x8000) RAM[A] = V;
return;
}

byte Z80_In(byte Port)
{
// printf("Blank In.\n");
return(0);
}

void Z80_Out(byte Port,byte Value)
{
// printf("Blank Out.\n");
return;
}

void Z80_Reti(void)
{
// printf("Blank Reti.\n");
return;
}

void Z80_Retn(void)
{
// printf("Blank Retn.\n");
return;
}

int Z80_Interrupt(void)
{
DipSwitchSYNC = 1;
if (kbd_in(25))
	{
	while (kbd_in(25)); // paused
	while (!kbd_in(25));
	while (kbd_in(25));  // pressing P again unpauses
	}

return(Z80_IGNORE_INT);
}

void Z80_Patch(Z80_Regs *Regs)
{ 
// printf("Blank Patch.\n");
return;
}

void Debug()
{
// printf("Blank Debugger.\n");
return;
}


